import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.sf.picard.filter.NotPrimaryAlignmentFilter;
import net.sf.picard.filter.SamRecordFilter;
import net.sf.picard.util.Interval;
import net.sf.picard.util.IntervalList;
import net.sf.picard.util.SamLocusIterator;
import net.sf.samtools.SAMFileHeader;
import net.sf.samtools.SAMFileReader;
import net.sf.samtools.SAMFileReader.ValidationStringency;

public class ClustersUtils {
	public static List<GeneEntry> gatherGenesGff3All(BufferedReader gffReader)
			throws IOException {
		String strLine;
		List<GeneEntry> genes = new ArrayList<GeneEntry>();
		GeneEntry currentGene = null;
		while ((strLine = gffReader.readLine()) != null) {
			strLine.trim();
			// split the line
			if (strLine.startsWith("#") == false) {
				String[] line = strLine.split("\t");
				// create gene object
				if (line[2].equalsIgnoreCase("gene")) {
					// add the last gene that was worked on
					if (currentGene != null) {
					// getting exons, merging exons, putting them back in
						List<ExonEntry> exL = currentGene.getExons();
					    List <ExonEntry> exLm = mergeExons(exL);
					    currentGene.setExons(exLm);
						genes.add(currentGene);
					}
					// start working on a new GeneEntry
					// get gene name and ID
					String[] geneName_ID = line[8].split(";");
					// store ID, name, alias and
					String geneID = "";
					String geneName = "";
					String geneAlias = "";
					String note = "";
					for (String s : geneName_ID) {
						if (s.trim().startsWith("ID=")) {
							geneID = s.substring(s.indexOf('=') + 1);
						}
						if (s.trim().startsWith("Name=")) {
							geneName = s.substring(s.indexOf('=') + 1);
						}
						if (s.trim().startsWith("Alias=")) {
							geneAlias = s.substring(s.indexOf('=') + 1);
						}
						if (s.trim().startsWith("Note=")) {
							note = s.substring(s.indexOf('=') + 1);
						}
					}
					// remember that they are indexed form 1
					String chromo = line[0].trim();
					Integer geneStart = Integer.parseInt(line[3]);
					Integer geneEnd = Integer.parseInt(line[4]);
					String orient = line[6];
					int or;
					if (orient.equals("+")) {
						or = 1;
					} else {
						if (orient.equals("-")) {
							or = -1;
						} else {
							or = 0;
						}
					}
					currentGene = new GeneEntry(geneID, geneStart, geneEnd, or,
							chromo, false);
					currentGene.setName(geneID);
					currentGene.setAlias(geneAlias);
					currentGene.setNote(note);
				} else {
					if (line[2].equalsIgnoreCase("mRNA") && currentGene != null) {
						String orient = line[6];
						int or;
						if (orient.equals("+")) {
							or = 1;
						} else {
							if (orient.equals("-")) {
								or = -1;
							} else {
								or = 0;
							}
						}
						// gather the information to create mRNA
						// will need something to add exon/CDS orintation to
						// mRNA objects, do not need it right now, but probably
						// will
						String[] mName_ID = line[8].split(";");
						// store ID, name, alias and
						String mID = "";
						String mName = "";
						String mAlias = "";
						for (String s : mName_ID) {
							if (s.trim().startsWith("ID=")) {
								mID = s.substring(s.indexOf('=') + 1);
							}
							if (s.trim().startsWith("Name=")) {
								mName = s.substring(s.indexOf('=') + 1);
							}
							if (s.trim().startsWith("Alias=")) {
								mAlias = s.substring(s.indexOf('=') + 1);
							}
						}
						String chromo = line[0].trim();
						Integer mRNAstart = Integer.parseInt(line[3]);
						Integer mRNAend = Integer.parseInt(line[4]);
						mRNAEntry m = new mRNAEntry(mID, mRNAstart, mRNAend, or);
						m.setName(mName);
						m.setAlias(mAlias);
						m.setChromosome(chromo);
						currentGene.addmRNA(m);
					}
					if (line[2].equalsIgnoreCase("exon") && currentGene != null) {
						String[] exonInfo = line[8].split(";");
						String exonIDFull = exonInfo[0];
						String exonID = exonIDFull.substring(exonIDFull
								.indexOf('=') + 1);
						Integer exonStart = Integer.parseInt(line[3]);
						Integer exonEnd = Integer.parseInt(line[4]);
						ExonEntry currentExon = new ExonEntry(exonID,
								exonStart, exonEnd);
						if (!currentGene.getExons().contains(currentExon)) {
							currentGene.addExon(currentExon);
						}
					}
				}
			}
		}
		// add the last gene
		if (currentGene != null) {
			// getting exons, merging exons, putting them back in
			List<ExonEntry> exL = currentGene.getExons();
		    List <ExonEntry> exLm = mergeExons(exL);
		    currentGene.setExons(exLm);
			genes.add(currentGene);
		}
		return genes;
	}

	// remeber exons are overlapping; remove the overlapping bits before
	// proceeding with checking
	// TODO !!!!!!!!!!!!!!!! remove exon overlaps
	public static void checkGenesUseExons(List<Integer> coverage,
			List<GeneEntry> geneList) {
		System.out.println("Gene list size is: " + geneList.size());
		int exonsLength;
		int exonsPosCovered;
		int totCoverageE;
		int genePosCovered;
		int geneLength;
		int totCoverageG;
		for (GeneEntry g : geneList) {
			genePosCovered = 0;
			geneLength = 0;
			totCoverageG = 0;
			for (int j = g.getStart() - 1; j < g.getEnd(); j++) {
				if (coverage.get(j) > 0) {
					genePosCovered++;
					totCoverageG=totCoverageG+coverage.get(j);
				}
				geneLength++;
			}
			double covF = (double) genePosCovered / (double) geneLength;
			g.setCovFrac(covF);
			double aveCovG=(double)totCoverageG/(double)genePosCovered;
			g.setAveGeneCov(aveCovG);
			exonsLength = 0;
			exonsPosCovered = 0;
			totCoverageE = 0;
			if (g.getExons() != null && g.getExons().size() != 0) {
				for (ExonEntry ex : g.getExons()) {
					// have to move all by one indexing from 0, positions from 1
					// for now only consider the positions over the desired
					// coverage in the average coverage count
					int exonSize = 0;
					int exonCov = 0;
					for (int i = ex.getStart() - 1; i < ex.getEnd(); i++) {
						if (coverage.get(i) > 0) {
							totCoverageE = totCoverageE + coverage.get(i);
							exonsPosCovered++;
							exonCov++;
						}
						exonsLength++;
						exonSize++;
					}
					ex.setFracCov((double) exonCov / (double) exonSize);
				}
				if (exonsLength != 0) {
					double covFrac = (double) exonsPosCovered
							/ (double) exonsLength;
					if (exonsPosCovered == 0) {
						g.setAveExonsCov(0.0);
					} else {
						double aveCov = (double) totCoverageE
								/ (double) exonsPosCovered;
						g.setAveExonsCov(aveCov);
					}
					g.setFeacExonsCov(covFrac);
				}
			}
		}
	}

	public static List<GeneEntry> getAllGenesGff3(String pathToGff)
			throws IOException {
		FileInputStream g1 = new FileInputStream(pathToGff);
		DataInputStream g2 = new DataInputStream(g1);
		BufferedReader g3 = new BufferedReader(new InputStreamReader(g2));
		return ClustersUtils.gatherGenesGff3All(g3);
	}

	public static Map<String, List<GeneEntry>> genesListToMap(
			List<GeneEntry> genes) {
		Map<String, List<GeneEntry>> genesInChr = new LinkedHashMap<String, List<GeneEntry>>();
		for (GeneEntry g : genes) {
			if (genesInChr.containsKey(g.getChromosome())) {
				genesInChr.get(g.getChromosome()).add(g);
			} else {
				genesInChr.put(g.getChromosome(), new ArrayList<GeneEntry>());
				genesInChr.get(g.getChromosome()).add(g);
			}
		}
		return genesInChr;
	}

	public static SAMFileReader getBamFileReader(String bamFilePath)
			throws Exception {
		File bamFile = new File(bamFilePath);
		File baiFile = new File(bamFile.getAbsolutePath() + ".bai");
		SAMFileReader sam = new SAMFileReader(bamFile, baiFile, true);
		sam.setValidationStringency(ValidationStringency.SILENT);
		if (!sam.hasIndex())
			throw new Exception("Missing index");
		return sam;
	}

	public static List<Integer> getCoverage(String[] bamFiles,
			String bamFilesPath, String chromosome) {
		List<Integer> finalCoverage = null;
		for (String bam : bamFiles) {
			SAMFileReader samR = null;
			try {
				samR = getBamFileReader("" + bamFilesPath.trim() + bam.trim());
			} catch (Exception e) {
				e.printStackTrace();
				System.exit(1);
			}
			SAMFileHeader headS = samR.getFileHeader();
			List<Integer> coverage = new ArrayList<Integer>();
			IntervalList seqIL = new IntervalList(headS);
			Interval seqI = new Interval(chromosome, 0, headS.getSequence(
					chromosome).getSequenceLength());
			seqIL.add(seqI);
			SamLocusIterator lItr = new SamLocusIterator(samR, seqIL, true);
			System.out.println("I'm chromosome: "
					+ headS.getSequence(chromosome).getSequenceName()
					+ " and I'm "
					+ headS.getSequence(chromosome).getSequenceLength()
					+ " bp long.");
			lItr.setEmitUncoveredLoci(true);
			// TODO testing, flter out not primary alignment
			//List<SamRecordFilter> filters = new ArrayList<SamRecordFilter>();
			//filters.add(new NotPrimaryAlignmentFilter());
			//lItr.setSamFilters(filters);
			lItr.setSamFilters(null);
			int lastPosition = 0;
			for (SamLocusIterator.LocusInfo inf : lItr) {
				int pos = inf.getPosition();
				if (pos == (lastPosition + 2)) {
					int cov = 0;
					coverage.add(cov);
				}
				coverage.add(inf.getRecordAndPositions().size());
				lastPosition = pos;
			}
			// remedy for the situation where there is a single 0 following
			// covered or uncovered position at the end, normally would not get
			// emitted
			// used to be a problem, not sure if it still is
			if (lastPosition == headS.getSequence(chromosome)
					.getSequenceLength() - 1) {
				coverage.add(0);
			}
			if (coverage.size() != headS.getSequence(chromosome)
					.getSequenceLength()) {
				System.out
						.println("WARNING: coverege array size not equals what reported in the header");
			}
			if (bamFiles.length == 1) {
				finalCoverage = coverage;
			} else {
				if (finalCoverage == null) {
					finalCoverage = new ArrayList<Integer>();
					for (int i = 0; i < headS.getSequence(chromosome)
							.getSequenceLength(); i++) {
						finalCoverage.add(0);
					}
				}
				for (int i = 0; i < coverage.size(); i++) {
					int cov = finalCoverage.get(i);
					int newCov = cov + coverage.get(i);
					finalCoverage.set(i, newCov);
				}
			}
			lItr.close();
			samR.close();
		}
		return finalCoverage;
	}

	public static Map<Integer, Set<Point>> putPointsToClusters(List<Point> pL) {
		Map<Integer, Set<Point>> clusters = new HashMap<Integer, Set<Point>>();
		for (Point p : pL) {
			int clu = p.getCluster();
			if (clusters.containsKey(clu)) {
				clusters.get(clu).add(p);
			} else {
				clusters.put(clu, new HashSet<Point>());
				clusters.get(clu).add(p);
			}
		}
		return clusters;
	}

	public static Map<String, Set<Point>> putPointsToClustersNames(
			List<Point> pL) {
		Map<String, Set<Point>> clusters = new HashMap<String, Set<Point>>();
		for (Point p : pL) {
			String clu = p.getClusterName();
			if (clusters.containsKey(clu)) {
				clusters.get(clu).add(p);
			} else {
				clusters.put(clu, new HashSet<Point>());
				clusters.get(clu).add(p);
			}
		}
		return clusters;
	}

	public static List<Cluster> mergeClusters(List<Cluster> cList, int diff) {

		List<Cluster> cListT = new ArrayList<Cluster>();
		cListT.add(0, cList.get(0));
		for (int i = 1; i < cList.size(); i++) {
			if (cList.get(i).getStart()
					- cListT.get(cListT.size() - 1).getEnd() - 1 < diff) {
				Cluster curr = cListT.get(cListT.size() - 1);
				curr.setEnd(cList.get(i).getEnd());
				curr.appendID(cList.get(i).getID());
				curr.addGenes(cList.get(i).getGeneNo());
			} else {
				cListT.add(cList.get(i));
			}
		}

		return cListT;
	}
	
	public static List<ExonEntry> mergeExons (List<ExonEntry> eList) {
		List<ExonEntry> eListM = new ArrayList<ExonEntry>();
		Collections.sort(eList, new ExonEntryComparator());
		eListM.add(0, eList.get(0));
		for (int i = 1; i < eList.size(); i++) {
			if (eList.get(i).getStart()
					<= eListM.get(eListM.size() - 1).getEnd()) {
				ExonEntry curr = eListM.get(eListM.size() - 1);
				if(eList.get(i).getEnd() > curr.getEnd()) {
				curr.setEnd(eList.get(i).getEnd());
				}
				curr.setID(curr.getID() + "_" + eList.get(i).getID());
			} else {
				eListM.add(eList.get(i));
			}
		}
		return eListM;
	}
}
