import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

public class DBSCAN {
	public static Set<Point> getNeighbors(Point p, List<Point> pL, double e) {
		Set<Point> neigh = new HashSet<Point>();
		for (Point q : pL) {
			if (Point.distance(p, q) <= e) {
				if (!p.equals(q)) {
					neigh.add(q);
				}
			}
		}
		return neigh;
	}

	public static void doDBSCAN(double e, int minClu, Set<Point> visited,
			List<Point> pL) {
		// visited is the vector with all the points that have been classified
		// as sth, noise or cluster
		int C = 0;
		for (Point p : pL) {
			if (visited.contains(p) == false) {
				visited.add(p);
				Set<Point> neighbors = getNeighbors(p, pL, e);
				if (neighbors.size() < minClu) {
					// mark as noise, do not do anything, already default to
					// noise null, mark it as -1
					p.setCluster(-1);
				} else {
					C++;
					expandCluster(p, neighbors, C, e, minClu, visited, pL);
				}
			}
		}
	}

	public static void mergeNeigbors(List<Point> n1, Set<Point> n2) {
		int cnt = 0;
		while (cnt < n1.size()) {
		}
	}

	public static void expandCluster(Point p, Set<Point> neighbors, int c,
			double e, int minC, Set<Point> visited, List<Point> pL) {
		p.setCluster(c);
		List<Point> neigh = new ArrayList<Point>(neighbors);
		if (neighbors.size() > 0) {
			int cnt = 0;
			while (cnt < neigh.size()) {
				//System.out.println("Count: " + cnt + " Neigh size: " + neigh.size());
				Point np = neigh.get(cnt);
				if (!visited.contains(np)) {
					visited.add(np);
					Set<Point> neighborsNp = getNeighbors(np, pL, e);
					if (neighborsNp.size() >= minC) {
						// may add the same object multiple times;
						// use a merge method instead
						// does not matter - visited solves the problem
						// even if the same point multiple times in the array
						// if maked as visited once, stays visited
						//for (Point x : neighborsNp) {
						//	if (!neigh.contains(x)) {
						//		neigh.add(x);
						//	}
						//}
						neigh.addAll(neighborsNp);
						//System.out.println("Added");
					}
				}
				if (np.getCluster() == null || np.getCluster() == -1) {
					np.setCluster(c);
				}
				cnt++;
			}
			neighbors = new HashSet<Point>(neigh);
		}
	}

	public static void main(String[] args) {
		// read points from file into array into array
		// as u create point objects constructor sets cluster to null
		List<Point> points = new ArrayList<Point>();
		Map<Integer, String> geneNames = new HashMap<Integer, String>();
		Set<Point> visited = new HashSet<Point>();
		// Point p1 = new Point(1);
		//
		// Point p2 = new Point(2);
		//
		// Point p3 = new Point(3);
		//
		// Point p4 = new Point(4);
		//
		// Point p5 = new Point(5);
		//
		// Point p6 = new Point(6);
		//
		// Point p7 = new Point(7);
		//
		// Point p8 = new Point(15);
		//
		// Point p9 = new Point(100);
		//
		// points.add(p1);
		//
		// points.add(p2);
		//
		// points.add(p3);
		//
		// points.add(p4);
		//
		// points.add(p5);
		//
		// points.add(p6);
		//
		// points.add(p7);
		//
		// points.add(p8);
		//
		// points.add(p9);
		//
		// sort the list based on p location
		// or read in list of sorted objects
		// find k-nearest neighbors based on 1D location of point
		//
		FileInputStream s1;
		DataInputStream s2;
		BufferedReader s3;
		try {
			s1 = new FileInputStream(
					"/home/agnieszka/BioInf/wheat-2BS/synteny/chromosomes_for_clustering/Bd1_gene_order_synteny.csv");
			s2 = new DataInputStream(s1);
			s3 = new BufferedReader(new InputStreamReader(s2));
			String strLine;
			while ((strLine = s3.readLine()) != null) {
				String[] arr = strLine.split(",");
				geneNames.put(Integer.parseInt(arr[1]), arr[0]);
				// adding end, 2D point now
				Point curP = new Point(Integer.parseInt(arr[1]));
				// curP.setLoc2(Integer.parseInt(arr[2]));
				points.add(curP);
			}
			s2.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
		int minClu = 12;
		double e = 10.00;
		doDBSCAN(e, minClu, visited, points);
		Map<Integer, Set<Point>> pc = ClustersUtils.putPointsToClusters(points);
		for (int i : pc.keySet()) {
			System.out.println("Cluster: " + i);
			for (Point p : pc.get(i)) {
				System.out.println(geneNames.get(p.getLoc()) + "," + p.getLoc()
						+ "," + i);
			}
			System.out.println();
		}
	}
}
