import java.io.FileOutputStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import net.sf.samtools.SAMFileHeader;
import net.sf.samtools.SAMFileReader;
import net.sf.samtools.SAMSequenceDictionary;
import net.sf.samtools.SAMSequenceRecord;

public class SyntenyClustersFinal {
	private static String bamPathKey = "bamPath";
	private static String gffKey = "gffFile";
	private static String outPathKey = "outDirPath";
	private static String separator = "=";
	private static String minCovKey = "expectCov";
	private static String minFracHKey = "minFracHor";
	private static String minFracVKey = "minFracVer";
	private static String chromListKey = "chromosomeList";
	private static String bamListKey = "bamFileList";
	private static String eKey = "DBepsilon";
	private static String minKey = "DBmin";
	private static String help = "";
	private static String gOe = "genesORexons";
	private static String mKey = "mergeDistance";
	private static String estKey = "estimateMinFracVer";
	
	public static String getHelp(String[] cmd) {
		for (String c : cmd) {
			if (c.equals("help")) {
				return help;
			}
		}
		return null;
	}

	public static Boolean getEst(String[] cmd) {
		for (String c : cmd) {
			if (c.equals("estimateMinFracVer")) {
				return true;
			}
		}
		return false;
	}
	public static Integer getDBEKey(String[] cmd) {
		Integer e = null;
		for (String c : cmd) {
			if (c.startsWith(eKey + separator)) {
				try {
					e = Integer.parseInt(c.substring(c.indexOf(separator) + 1));
					System.out.println(Integer.parseInt(c.substring(c
							.indexOf(separator) + 1)));
				} catch (NumberFormatException nfe) {
					return null;
				}
				return e;
			}
		}
		return null;
	}

	public static Integer getMergeDistance(String[] cmd) {
		Integer m = null;
		for (String c : cmd) {
			if (c.startsWith(mKey + separator)) {
				try {
					m = Integer.parseInt(c.substring(c.indexOf(separator) + 1));
					System.out.println(Integer.parseInt(c.substring(c
							.indexOf(separator) + 1)));
				} catch (NumberFormatException nfe) {
					return null;
				}
				return m;
			}
		}
		return null;
	}

	public static Integer getDBMinKey(String[] cmd) {
		Integer m = null;
		for (String c : cmd) {
			if (c.startsWith(minKey + separator)) {
				try {
					m = Integer.parseInt(c.substring(c.indexOf(separator) + 1));
					System.out.println(Integer.parseInt(c.substring(c
							.indexOf(separator) + 1)));
				} catch (NumberFormatException e) {
					return null;
				}
				return m;
			}
		}
		return null;
	}

	public static String getBamFileList(String[] cmd, String key) {
		for (String c : cmd) {
			if (c.startsWith(key + separator)) {
				System.out.println(c.substring(c.indexOf(separator) + 1));
				return c.substring(c.indexOf(separator) + 1);
			}
		}
		return null;
	}

	public static String getChromList(String[] cmd) {
		for (String c : cmd) {
			if (c.startsWith(chromListKey + separator)) {
				System.out.println(c.substring(c.indexOf(separator) + 1));
				return c.substring(c.indexOf(separator) + 1);
			}
		}
		return null;
	}

	public static String getBamPath(String[] cmd, String key) {
		for (String c : cmd) {
			if (c.startsWith(key + separator)) {
				System.out.println(c.substring(c.indexOf(separator) + 1));
				return c.substring(c.indexOf(separator) + 1);
			}
		}
		return null;
	}

	public static String getGenesExons(String[] cmd) {
		for (String c : cmd) {
			if (c.startsWith(gOe + separator)) {
				System.out.println(c.substring(c.indexOf(separator) + 1));
				return c.substring(c.indexOf(separator) + 1);
			}
		}
		return null;
	}

	public static Double getExpectCov(String[] cmd) {
		Double exp = null;
		for (String c : cmd) {
			if (c.startsWith(minCovKey + separator)) {
				try {
					exp = Double
							.parseDouble(c.substring(c.indexOf(separator) + 1));
					System.out.println(Double.parseDouble(c.substring(c
							.indexOf(separator) + 1)));
				} catch (NumberFormatException e) {
					return null;
				}
				return exp;
			}
		}
		return null;
	}

	public static Double getMinFracH(String[] cmd) {
		Double mf = null;
		for (String c : cmd) {
			if (c.startsWith(minFracHKey + separator)) {
				try {
					mf = Double
							.parseDouble(c.substring(c.indexOf(separator) + 1));
					System.out.println(Double.parseDouble(c.substring(c
							.indexOf(separator) + 1)));
				} catch (NumberFormatException e) {
					return null;
				}
				return mf;
			}
		}
		return null;
	}

	public static Double getMinFracV(String[] cmd) {
		Double mf = null;
		for (String c : cmd) {
			if (c.startsWith(minFracVKey + separator)) {
				try {
					mf = Double
							.parseDouble(c.substring(c.indexOf(separator) + 1));
					System.out.println(Double.parseDouble(c.substring(c
							.indexOf(separator) + 1)));
				} catch (NumberFormatException e) {
					return null;
				}
				return mf;
			}
		}
		return null;
	}
	public static String getOutDirPath(String[] cmd) {
		for (String c : cmd) {
			if (c.startsWith(outPathKey + separator)) {
				System.out.println(c.substring(c.indexOf(separator) + 1));
				return c.substring(c.indexOf(separator) + 1);
			}
		}
		return null;
	}

	public static String getGffFile(String[] cmd) {
		for (String c : cmd) {
			if (c.startsWith(gffKey + separator)) {
				System.out.println(c.substring(c.indexOf(separator) + 1));
				return c.substring(c.indexOf(separator) + 1);
			}
		}
		return null;
	}

	public static void main(String[] args) {
		String h = getHelp(args);
		if (h != null) {
			System.out.println(help);
			System.exit(1);
		}
		if (args.length < 5) {
			System.out.println(help);
			System.exit(1);
		}
		String bamFilesPath = getBamPath(args, bamPathKey);
		if (bamFilesPath == null) {
			System.out.println("Path to bam files not specified.");
			System.out.println(help);
			System.exit(1);
		}
		String bamFiles = getBamFileList(args, bamListKey);
		String[] bamFilesL=bamFiles.split(",");
		if (bamFilesL == null || bamFilesL.length == 0) {
			System.out.println("List of .bam files not specified.");
			System.out.println(help);
			System.exit(1);
		}
		Boolean estV = getEst(args);
		
		Double expCov = getExpectCov(args);
		if (expCov == null && estV==false) {
			System.out
					.println("Expected coverage not specified.");
			System.out.println(help);
			System.exit(1);
		}
		else {
			if(expCov == null && estV==false) {
				System.out
				.println("You have asked for vertical coverage cutoff estimation and specified expectCov.");
				System.out.println("expectCov will be ignored.");
			}
		}
		// set to zero, want everything right now
		Double synCutOffH = getMinFracH(args);
		if (synCutOffH == null) {
			System.out
					.println("Synteny cutoff is not specified or not within range 0.0<=x<=1.0.");
			System.out.println("Using default value 0.3");
			synCutOffH=0.3;
		}
		
		Double synCutOffV = getMinFracV(args);
		if (synCutOffV == null && estV==false) {
			System.out
					.println("Synteny cutoff is not specified or not within range 0.0<=x<=1.0.");
			System.out.println("Using default value 0.05");
			synCutOffV=0.05;
		}
		else {
			if(synCutOffV == null && estV==true) {
				System.out
				.println("You have asked for vertical coverage cutoff estimation and specified minFracVer.");
				System.out.println("minFracVer will be ignored.");
			}
		}
		String gffPath = getGffFile(args);
		if (gffPath == null) {
			System.out.println("Gff3 file not specified.");
			System.out.println(help);
			System.exit(1);
		}
		String outPath = getOutDirPath(args);
		if (outPath == null) {
			System.out.println("Path to the output folder not specified.");
			System.out.println(help);
			System.exit(1);
		}
		String chromList = getChromList(args);
		if (chromList == null) {
			System.out.println("Chromosome list not specified.");
			System.out.println("Using default value: all");
			chromList="all";
		}
		String[] chrs = chromList.trim().split(",");
		if (chrs.length == 0) {
			System.out.println("Chromosome list empty.");
			System.out.println(help);
			System.exit(1);
		}
		Integer minClu1 = getDBMinKey(args);
		if (minClu1 == null) {
			System.out.println("DBMin is not specified or not an integer");
			System.out.println("Using default value: 24");
			minClu1=24;
		}
		Integer eps1 = getDBEKey(args);
		if (eps1 == null) {
			System.out.println("DBEpsilon is not specified or not an integer");
			System.out.println("Using default value: 26");
			eps1=26;
		}
		String genesExons = getGenesExons(args);
		if (genesExons == null) {
			System.out.println("Genes or Exons not specified");
			System.out.println("Using default value: exons");
			genesExons="exons";
		}
		Boolean exons = null;
		if (genesExons.equalsIgnoreCase("exons")) {
			exons = true;
			System.out.println("Exons");
		} else {
			if (genesExons.equalsIgnoreCase("genes")) {
				exons = false;
				System.out.println("Genes");
			} else {
				System.out.println("Wrong genes or exons value");
				System.out.println("Using default value: exons");
				genesExons="exons";
			}
		}
		Integer dist = getMergeDistance(args);
		if (dist == null || dist < 1) {
			System.out.println("No distance specified or distance < 1");
			System.out.println("Using default value: 20");
			dist=20;
		}
		PrintWriter print1 = null;
		try {
			print1 = new PrintWriter(new FileOutputStream(outPath + "stats.txt", false));
			print1.println("Bam path: " + bamFilesPath);
			print1.println("Bam files: " + bamFiles);
			print1.println("Expected cov: " + expCov);
			print1.println("Synteny cutoff horizontal: " + synCutOffH);
			print1.println("Synteny cutoff vertical: " + synCutOffV);
			print1.println("Gff path: " + gffPath);
			print1.println("Chromosome list: " + chromList);
			print1.println("Min cluster size: " + minClu1);
			print1.println("Radius: " + eps1);
			print1.println("Gene or exons: " + genesExons);
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		if (chrs[0].equalsIgnoreCase("all")) {
			SAMFileReader samR = null;
			try {
				samR = ClustersUtils.getBamFileReader("" + bamFilesPath.trim()
						+ bamFilesL[0]);
			} catch (Exception e) {
				e.printStackTrace();
				System.exit(1);
			}
			SAMFileHeader headS = samR.getFileHeader();
			SAMSequenceDictionary dictS = headS.getSequenceDictionary();
			List<SAMSequenceRecord> seqsR = dictS.getSequences();
			String[] chrArr = new String[seqsR.size()];
			int cnt = 0;
			for (SAMSequenceRecord s : seqsR) {
				chrArr[cnt] = s.getSequenceName();
				cnt++;
			}
			samR.close();
			chrs = chrArr;
		}
		// part I - get genes from gff3 file
		List<GeneEntry> geneList = null;
		try {
			geneList = ClustersUtils.getAllGenesGff3(gffPath);
		} catch (Exception e) {
			e.printStackTrace();
		}
		Map<String, List<GeneEntry>> genesInChrs = ClustersUtils
				.genesListToMap(geneList);
		for (String s : chrs) {
			if (genesInChrs.get(s) == null) {
				continue;
			}
			List<Integer> finalCoverage = ClustersUtils.getCoverage(bamFilesL,
					bamFilesPath, s);
			ClustersUtils.checkGenesUseExons(finalCoverage, genesInChrs.get(s));
		}
		// sorting the genes in chromosomes by start positions
		for (String chr : genesInChrs.keySet()) {
			if (genesInChrs.get(chr) == null) {
				continue;
			}
			Collections.sort(genesInChrs.get(chr), new GeneEntryComparator());
		}
		for (String chr : genesInChrs.keySet()) {
			if (genesInChrs.get(chr) == null) {
				continue;
			}
			
			//trying to evaluate coverage distribiution
			Map<Integer,Integer> covDist = new TreeMap<Integer,Integer>();
			for (int i = 0; i < genesInChrs.get(chr).size(); i++) {
				GeneEntry gene1 = genesInChrs.get(chr).get(i);
				double covFracH;
				if(exons) {
					covFracH=gene1.getAveExonsCov();
				}
				else {
					covFracH=gene1.getAveGeneCov();
				}
				if(covFracH <= synCutOffH || covFracH >= 200) {
					continue;
				}
				double covFracV;
				if (exons) {
					covFracV = gene1.getAveExonsCov();
				} else {
					covFracV = gene1.getAveGeneCov();
				}
				int covFracVInt=(int) Math.round(covFracV);
				if(covDist.containsKey(covFracVInt)) {
					covDist.put(covFracVInt, covDist.get(covFracVInt) + 1);
				}
				else {
					covDist.put(covFracVInt, 1);
				}	
			}
			// trying to coverage distribution to estimate
			int minCovVEst=0;
			int geneNo=0;
			for (int i : covDist.keySet() ) {
				if(i != 0) {
					geneNo=geneNo+covDist.get(i);
				}
			}
			int geneSum=0;
			for (int i : covDist.keySet() ) {
				if(i != 0) {
					geneSum=geneSum+covDist.get(i);
					if(geneSum >= 0.8*geneNo) {
						minCovVEst=i;
						break;
					}
				}
			}
			System.out.println("Gene sum: " + geneSum);
			System.out.println("Gene no: " + geneNo);
			System.out.println("Cutoff: " + minCovVEst );
			// cutoff for vertical coverage value
			double minCovV;
			if(estV) {
				minCovV=minCovVEst;
			}
			else {
				minCovV= synCutOffV * expCov;
			}
			print1.println("Min coverage vertical is: " + minCovV);
			// printing to the stats file
			// always add one to make indexing from one
			// points and names have to be done separately for each
			// chromosome!!!!!!!
			// update use the synCutOff to choose points for clustering
			//System.out.println("Chromosome: " + chr);
			Map<String, Point> points1Map = new HashMap<String, Point>();
			List<Point> points1 = new ArrayList<Point>();
			Set<Point> visited1 = new HashSet<Point>();
			double covVcutoff=200;
			if(expCov != null) {
				covVcutoff=expCov;
			}
			
			for (int i = 0; i < genesInChrs.get(chr).size(); i++) {
				GeneEntry gene1 = genesInChrs.get(chr).get(i);
				// fixed a bug was looking at coverage of whole genes not exons
				// change - looking at vertical coverage
				double covFracH = 0;
				double covV=0;
				if (exons) {
					covV = gene1.getAveExonsCov();
					covFracH= gene1.getFracExonsCov();
				} else {
					covV = gene1.getAveGeneCov();
					covFracH = gene1.getCovFrac();
				}
				if (covFracH > synCutOffH && covV > minCovV && covV <= covVcutoff) {
					Point point1 = new Point(i + 1);
					points1.add(point1);
					points1Map.put(gene1.getID(), point1);
				}

			}
			DBSCAN.doDBSCAN(eps1, minClu1, visited1, points1);
			Map<Integer, Set<Point>> pc = ClustersUtils
					.putPointsToClusters(points1);
			Map<Integer, List<Point>> sortedPc = new TreeMap<Integer, List<Point>>();
			Map<Integer, Double> clusterScores = new HashMap<Integer, Double>();
			// no point is less than zero
			int biggestCluster = 0;
			List<Cluster> cList = new ArrayList<Cluster>();
			for (int clust : pc.keySet()) {
				List<Point> pList = new ArrayList<Point>(pc.get(clust));
				if (pList.size() > biggestCluster) {
					biggestCluster = pList.size();
				}
				Collections.sort(pList, new PointComparator());
				sortedPc.put(clust, pList);
				int min = pList.get(0).getLoc();
				int max = pList.get(pList.size() - 1).getLoc();
				// exclude -1
				if (clust != -1) {
					//System.out.println(clust);
					cList.add(new Cluster(clust + "", min, max, pList.size()));
				}
				Collections.sort(cList, new ClusterComparator());
				// score according to gene density
				double score = (double) pList.size() / (double) (max - min + 1);
				clusterScores.put(clust, score);
				// System.out.println(clust + " " + score);
				// System.out.println(pList.size());
				// System.out.println(max - min + 1);
			}
			List<Cluster> mergedC = null;
			if (cList.size() > 1) {
				mergedC = ClustersUtils.mergeClusters(cList, dist);
				// now need to reassign points to new clusters
				Map<Integer, String> cN = new HashMap<Integer, String>();
				for (Cluster c : mergedC) {
					String id = c.getID();
					String[] id_arr = id.split("_");
					for (String i : id_arr) {
						cN.put(Integer.parseInt(i), id);
					}
				}
				for (Point p : points1) {
					p.setClusterName(cN.get(p.getCluster()));
				}

			}
			PrintWriter print2 = null;
			try {
				print2 = new PrintWriter(new FileOutputStream(outPath + chr
						+ ".clusters", false));
				print2.println("chromosome,start_pos,gene_name,no_of_exons,cov_fraction,cov_depth,cluster,merged_clusters");
				for (GeneEntry gene : genesInChrs.get(chr)) {
					double cov = 0;
					if (exons) {
						cov = gene.getFracExonsCov();
					} else {
						cov = gene.getCovFrac();
					}
					if (points1Map.containsKey(gene.getID())) {
						int clust = points1Map.get(gene.getID()).getCluster();
						String clustM;
						if (cList.size() > 1) {
							if(points1Map.get(gene.getID()).getCluster()!=-1) {
							clustM = points1Map.get(gene.getID())
									.getClusterName();
							}
							else {
								clustM="-1";
							}
						} else {
							clustM = clust + "";
						}
						print2.println(chr + "," + gene.getStart() + ","
								+ gene.getID() + "," + gene.getExons().size()
								+ "," + cov + "," + gene.getAveExonsCov() + "," +"Cluster_" + clust + "," 
								+ "ClusterM_" + clustM);
					} else {
						print2.println(chr + "," + gene.getStart() + ","
								+ gene.getID() + "," + gene.getExons().size()
								+ "," + cov + "," + gene.getAveExonsCov() + "," + "NC" + "," + "NC");
					}
				}
				print2.close();
			} catch (Exception e) {
				e.printStackTrace();
			}
			print1.println(chr);
			for (int clust : clusterScores.keySet()) {
				print1.println(clust + ": " + clusterScores.get(clust));
			}

			if (cList.size() > 1) {
				for (Cluster c : mergedC) {
					double score = (double) c.getGeneNo()
							/ (c.getEnd() - c.getStart() + 1);
					print1.println(c.getID() + ": " + score);
				}
			}
			
			print1.println("Coverage distribution");
			for (Integer i : covDist.keySet()) {
				print1.println(i + ": " + covDist.get(i));
			}
			
			//print1.println("Estimated min cov is: " + minCovVEst);
		}
		
		print1.close();
	}
}
